"""Contains a class holding current user settings."""
import dataclasses
from contextlib import suppress
from typing import Optional

from activate import serialise, units
from activate.app import connect, paths

DEFAULTS = {
    "tiles": None,
    "map_speed": 0.0,
    "unit_system": units.DEFAULT,
    "servers": [],
    "custom_units": {},
    "cookie": None,
}


def load_settings():
    """Load settings from a configuration file."""
    settings_data = DEFAULTS.copy()
    with suppress(FileNotFoundError):
        settings_data.update(serialise.load(paths.SETTINGS))

    settings_data["servers"] = [connect.Server(**s) for s in settings_data["servers"]]
    return Settings(**settings_data)


@dataclasses.dataclass
class Settings:
    """A settings configuration."""

    tiles: Optional[str]
    map_speed: float
    unit_system: str
    custom_units: dict
    servers: list
    cookie: Optional[str]

    def save(self):
        """Save settings to a configuration file."""
        serialise.dump(dataclasses.asdict(self), paths.SETTINGS, readable=True)

    def copy_from(self, other):
        for field in dataclasses.fields(self):
            setattr(self, field.name, getattr(other, field.name))



import re
import shutil
from collections import defaultdict
from dataclasses import dataclass
from typing import Optional
from uuid import UUID

from activate import serialise
from activate.app import files, load_activity, paths
from activate.app.sync import strava

FILENAME = re.compile(r'filename="(.*)"')


def download_response(response, path):
    try:
        filename = next(
            FILENAME.finditer(response.headers["Content-Disposition"])
        ).group(1)
    except KeyError:
        filename = response.url.split("/")[-1]
    filename = files.encode_name(filename, path)
    with open(filename, "wb") as f:
        response.raw.decode_content = True
        shutil.copyfileobj(response.raw, f)
    return filename


def add_photo_from_response(response, activity):
    filename = download_response(response, paths.PHOTOS)
    activity.photos.append(filename)


def import_from_response(response):
    filename = download_response(response, paths.TRACKS)
    activity = load_activity.load(filename)
    activity["filename"] = filename
    return activity


@dataclass
class SyncState:
    state: Optional[defaultdict] = None

    def ensure_loaded(self):
        if self.state is not None:
            return
        try:
            self.state = defaultdict(
                dict, {UUID(k): v for k, v in serialise.load(paths.SYNC_STATE).items()}
            )
        except FileNotFoundError:
            self.state = defaultdict(dict)

    def write(self):
        serialise.dump(
            {str(k): v for k, v in self.state.items()}, paths.SYNC_STATE, gz=True
        )

    def add(self, service, activity_id, service_activity_id):
        self.state[activity_id] = {service: service_activity_id}

    def get(self, service, activity_id):
        return self.state[activity_id].get(service)

    def sync(self, cookies, activities):
        """
        Download new activities.

        Returns a generator. The first value is the number of items after
        it. The other items are Activities.
        """
        yield from strava.sync_new(cookies["Strava"], activities, self)


sync_state = SyncState()



"""Sync activities with other services."""

import json
import re
from datetime import datetime

import requests
from bs4 import BeautifulSoup

import activate.activity
from activate.app import load_activity, sync

EFFORT_LEVEL_RE = re.compile(r"perceivedExertion&quot;:(\d+)\.0,")


def strava_url(page: str) -> str:
    return f"https://www.strava.com/{page}"


def get_strava_html(page: str, cookie: str) -> str:
    response = requests.get(strava_url(page), cookies={"_strava4_session": cookie})
    return response.text


def list_activities(cookie: str) -> dict:
    response = requests.get(
        strava_url("athlete/training?page=1"),
        headers={"X-Requested-With": "XMLHttpRequest"},
        cookies={"_strava4_session": cookie},
    )
    return {a["id"]: a for a in json.loads(response.text)["models"]}


def get_activity_data(strava_activity_id: int, cookie: str) -> requests.models.Response:
    page = f"activities/{strava_activity_id}"

    activity_file = requests.get(
        strava_url(f"{page}/export_original"),
        cookies={"_strava4_session": cookie},
        stream=True,
    )
    return activity_file


def get_edit_page_data(strava_activity_id: int, cookie: str) -> tuple:
    html = get_strava_html(f"activities/{strava_activity_id}/edit", cookie)
    photo_urls = [
        i.img["src"]
        for i in BeautifulSoup(html, "html.parser").find_all("div", "image-wrap")
    ]
    effort_level = EFFORT_LEVEL_RE.search(html)
    if effort_level is not None:
        effort_level = int(effort_level.group(1)) - 1

    return photo_urls, effort_level


def sport(data):
    return load_activity.convert_activity_type(data["type"], data["name"])


def update_local(activity, data):
    activity.name = data["name"]
    activity.sport = sport(data)
    activity.flags["Commute"] = data["commute"]
    activity.flags["Indoor"] = data["trainer"]
    if activity.sport == "Run":
        activity.flags["Treadmill"] = data["trainer"]
    activity.flags["Race"] = data["workout_type"] in {1, 11}
    activity.flags["Long Run"] = data["workout_type"] == 2
    activity.flags["Workout"] = data["workout_type"] in {3, 12}
    activity.description = (
        data["description"] if data["description"] is not None else ""
    )


def matches(activity, data):
    return all(
        (
            datetime.fromisoformat(data["start_time"][:-5]) == activity.start_time,
            sport(data) == activity.sport,
        )
    )


def sync_new(cookie, activities, sync_list):
    """
    Download new activities from Strava.

    Returns a generator. The first value is the number of items after
    it. The other items are Activities.
    """
    remote_activities = list_activities(cookie)
    for activity in activities:
        strava_id = sync_list.get("Strava", activity.activity_id)
        if strava_id is None:
            for remote_id, data in remote_activities.items():
                if matches(activity, data):
                    sync_list.add("Strava", activity.activity_id, remote_id)
                    strava_id = remote_id
                    break
            else:
                continue

        elif strava_id not in remote_activities:
            continue
        activity = activities.get_activity(activity.activity_id)
        update_local(activity, remote_activities[strava_id])
        activities.update(activity.activity_id)
        activities.save_activity(activity.activity_id)
        del remote_activities[strava_id]

    yield len(remote_activities)

    for strava_activity_id, data in remote_activities.items():
        activity = activate.activity.from_track(
            **sync.import_from_response(get_activity_data(strava_activity_id, cookie))
        )
        update_local(activity, data)
        photo_urls, effort_level = get_edit_page_data(strava_activity_id, cookie)
        activity.effort_level = effort_level
        for url in photo_urls:
            sync.add_photo_from_response(requests.get(url, stream=True), activity)

        sync_list.add("Strava", activity.activity_id, strava_activity_id)
        yield activity
